from enum import Enum


# from .formatter import Formatter


class Position(Enum):
    LEFT = 'left'
    Right = 'right'
    TOP = 'top'
    BELOW = 'below'


class Comment():
    def __init__(self, text: str, position: Position):
        self.text = text
        self.position = position


class Syntax():
    def __init__(self, value, comments: list):
        self.value = value
        self.comments = comments


class Package():
    def __init__(self, value, comments: list):
        self.value = value
        self.comments = comments


class Import():
    def __init__(self, value, comments: list):
        self.value = value
        self.comments = comments


class Option():
    def __init__(self, name, value, comments: list):
        self.name = name
        self.value = value
        self.comments = comments


class MessageElement():
    def __init__(self, type, name, number, rules='', label='', comments=[]):
        self.label = label
        self.type = type
        self.name = name
        self.number = number
        self.rules = rules
        self.comments = comments


class Message():
    def __init__(self, name, elements=[], comments=[]):
        self.name = name
        self.elements = elements
        self.comments = comments


class OneofElement():
    def __init__(self, type, name, number, rules='', label='', comments=[]):
        self.label = label
        self.type = type
        self.name = name
        self.number = number
        self.rules = rules
        self.comments = comments


class Oneof():
    def __init__(self, name, elements=[], comments=[]):
        self.name = name
        self.elements = elements
        self.comments = comments


class EnumElement():
    def __init__(self, name, number, rules='', comments=[]):
        self.name = name
        self.number = number
        self.rules = rules
        self.comments = comments


class ProtoEnum():
    def __init__(self, name, elements=[], comments=[]):
        self.name = name
        self.elements = elements
        self.comments = comments


#  rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
class ServiceElement():
    def __init__(self, label, name, request, response, comments=[]):
        self.label = label
        self.name = name
        self.request = request
        self.response = response
        self.comments = comments


class Service():
    def __init__(self, name, elements=[], comments=[]):
        self.name = name
        self.elements = elements
        self.comments = comments



from .proto_structures import Comment
from .comment import CommentParser
from .constant import Constant
from .proto_structures import Position
from .util import remove_prefix, remove_suffix


class Detector(Constant):

    def get_top_comments(self, lines):
        comment_parser = CommentParser()
        comments_lines = comment_parser.pick_up_comment(lines)
        comments = comment_parser.parse(comments_lines)

        return comments

    def get_type(self, line):
        if self._is_syntax_line(line):
            return 'syntax'
        if self._is_package_line(line):
            return 'package'
        if self._is_option_line(line):
            return 'option'
        if self._is_import_line(line):
            return 'import'
        if self._is_message_object(line):
            return 'message'
        if self._is_enum_object(line):
            return 'enum'
        if self._is_service_object(line):
            return 'service'
        # if self._is_element_line(line):
        #     return 'element_field', comments
        if self._is_message_element(line):
            return 'message_element'
        if self._is_enum_element(line):
            return 'enum_element'
        if self._is_service_element(line):
            return 'service_element'
        if self._is_oneof_object(line):
            return 'oneof'

        return 'unknown'

    def is_object_ype(self, type):
        return type in ['message', 'enum', 'service', 'oneof']

    def _is_syntax_line(self, line):
        return line.replace(' ', '').startswith('syntax=')

    def _is_package_line(self, line):
        return line.strip().startswith('package ')

    def _is_option_line(self, line):
        return line.strip().startswith('option ')

    def _is_import_line(self, line):
        return line.strip().startswith('import ')

    def _is_object_start(self, line):
        if line.count(self.BRACE_LEFT) == 0:
            return False

        if line.count(self.SINGLE_COMMENT_SYMBOL) > 0:
            if line.index(self.BRACE_LEFT) > line.index(self.SINGLE_COMMENT_SYMBOL):
                return False

        if line.count(self.MULTIPLE_COMENT_START_SYMBOL) > 0:
            if line.index(self.BRACE_LEFT) > line.index(self.MULTIPLE_COMENT_START_SYMBOL):
                return False

        return True

    def _is_object_end(self, line):
        if line.count(self.BRACE_RIGHT) == 0:
            return False

        if line.count(self.SINGLE_COMMENT_SYMBOL) > 0:
            if line.index(self.BRACE_RIGHT) > line.index(self.SINGLE_COMMENT_SYMBOL):
                return False

        if line.count(self.MULTIPLE_COMENT_START_SYMBOL) > 0:
            if line.index(self.BRACE_RIGHT) > line.index(self.MULTIPLE_COMENT_START_SYMBOL):
                return False

        return True

    def _is_message_object(self, line):
        return line.strip().startswith('message ') and line.strip().count(self.BRACE_LEFT)

    def _is_oneof_object(self, line):
        return line.strip().startswith('oneof ') and line.strip().count(self.BRACE_LEFT)

    def _is_message_element(self, line):
        if not self._is_element_line(line):
            return False

        if self._is_map_element(line):
            return True

        parts = line.strip().split(self.EQUAL_SIGN)
        parts = [e for e in parts[0].strip().split(' ')]
        parts = list(filter(None, parts))

        if parts[0] == 'rpc':  # service element
            return False

        return len(parts) >= 2

    def _is_element_line(self, line):
        if line.count(self.SEMICOLON) == 0:
            return False

        if line.count(self.SINGLE_COMMENT_SYMBOL) > 0:
            if line.index(self.SEMICOLON) > line.index(self.SINGLE_COMMENT_SYMBOL):
                return False

        if line.count(self.MULTIPLE_COMENT_START_SYMBOL) > 0:
            if line.index(self.SEMICOLON) > line.index(self.MULTIPLE_COMENT_START_SYMBOL):
                return False

        if self._is_service_element_line(line):
            return True

        if self._is_map_element_line(line):
            return True

        return line.strip().count(self.SEMICOLON) > 0 and line.strip().count(self.EQUAL_SIGN) > 0

    def _is_service_element(self, line):
        if not self._is_element_line(line):
            return False

        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = line.strip()
        return line.startswith('rpc ')

    def _is_service_element_line(self, line):
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = line.strip()
        return line.startswith('rpc ')

    def _is_map_element(self, line):
        if not self._is_element_line(line):
            return False

        # map<string, Project> projects = 3;
        line = line.strip().replace(' ', '')
        return line.startswith('map<')

    def _is_map_element_line(self, line):
        # map<string, Project> projects = 3;
        line = line.strip().replace(' ', '')
        return line.startswith('map<')

    def _is_enum_object(self, line):
        return line.strip().startswith('enum ') and line.strip().count(self.BRACE_LEFT)

    def _is_enum_element(self, line):
        if not self._is_element_line(line):
            return False

        parts = line.strip().split(self.EQUAL_SIGN)
        parts = [e for e in parts[0].strip().split(' ')]
        parts = list(filter(None, parts))

        return len(parts) == 1

    def _is_service_object(self, line):
        return line.strip().startswith('service ') and line.strip().count(self.BRACE_LEFT)



from src.proto_formatter import format_str

proto_str = """
    /*
    * *
                 *Person balabala
        **         asdas
        
*/
/* I'am a comment */
    message Person {
       string locale            = 3 [(validate.rules).string.pattern = "^[a-z]{2}-[A-Z]{2}$"];   // ISO 639-2 language code & ISO 3166-1 country code. e.g. en-US, en-GB
}
"""
formatted_proto_str = format_str(proto_str, align_by_equal_sign=True, comment_max_length=120)
print(formatted_proto_str)


import os
import sys
import types
import argparse
from proto_formatter import format_file
from .util import read_file


def _get_proto_files(root_path):
    proto_files = []
    for path, subdirs, files in os.walk(root_path):
        for name in files:
            if name.endswith('.proto'):
                proto_files.append(os.path.join(path, name))
    return proto_files


def get_option_max_length(parser):
    actions = parser._actions
    max_length = 0
    for action in actions:
        for option_string in action.option_strings:
            if max_length < len(option_string):
                max_length = len(option_string)
        if action.choices:
            for parser_name, parser in action.choices.items():
                if max_length < len(parser_name):
                    max_length = len(parser_name)
                for option_string_action in parser._option_string_actions:
                    if max_length < len(option_string_action):
                        max_length = len(option_string_action)

    return max_length


def format_actions(max_length_of_options, actions):
    saboah = 4  # space amount between option and help

    lines = ['general options:']
    for action in actions:
        if action.option_strings:
            fill_spaces_amount = max_length_of_options - len(action.option_strings[0]) + saboah
            line = f"{' ' * 2}{action.option_strings[0]}{' ' * fill_spaces_amount}{action.help}"
            lines.append(line)

    return lines


def make_usage(parser):
    max_length_of_names = get_option_max_length(parser)
    ONE_SPACE = ' '
    indents = 2
    saboah = 4  # space amount between option and help
    option_lines = ["", "general options:"]
    actions = parser._actions
    commands_lines = []
    for action in actions:
        for option_string in action.option_strings:
            fill_spaces_amount = max_length_of_names - len(option_string) + saboah
            line = f"{ONE_SPACE * indents}{option_string}{ONE_SPACE * fill_spaces_amount}{action.help}"
            option_lines.append(line)

        if action.choices:
            for command, sub_parser in action.choices.items():
                fill_spaces_amount = max_length_of_names - len(command) + saboah
                line = f"{ONE_SPACE * indents}{command}{ONE_SPACE * fill_spaces_amount}{sub_parser.description}"
                commands_lines.append(line)
                for action_name, action_value in sub_parser._option_string_actions.items():
                    for option_string in action_value.option_strings:
                        fill_spaces_amount = max_length_of_names - len(option_string) + saboah
                        line = f"{ONE_SPACE * indents}{option_string}{ONE_SPACE * fill_spaces_amount}{action_value.help}"
                        option_lines.append(line)
    content = """
{before_stars}
*  {description}  *
{end_stars}

usage:
  proto_formatter <command> [options]

commands:
"""
    before_stars = "*" * (len(parser.description) + 6)
    end_stars = before_stars
    content = content.format(description=parser.description, before_stars=before_stars, end_stars=end_stars)
    content = content + '\n'.join(commands_lines + option_lines)
    return content


def new_format_usage(self):
    formatter = self._get_formatter()
    # usage
    formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)

    lines = formatter.format_help().split('\n')
    lines[0] = ''  # replace the prefix "usage:"
    return '\n'.join(lines)


def new_format_help(self):
    formatter = self._get_formatter()
    # usage
    formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)

    lines = formatter.format_help().split('\n')
    lines[0] = ''  # replace the prefix "usage:"
    return '\n'.join(lines)


def main():
    parser = argparse.ArgumentParser(description="Format protobuf file(s) from a specific target.")  # add_help=False,
    sub_parser = parser.add_subparsers(dest='command')
    format_parser = sub_parser.add_parser('format', description='format protobuf files', add_help=False)
    view_parser = sub_parser.add_parser('view', description="view file", add_help=False)
    view_parser.add_argument(
        "--file",
        type=str,
        help="target protobuf file, only used for command 'view'"
    )
    format_parser.add_argument(
        "--files",
        type=str,
        nargs="+",
        help="target protobuf files need to be formatted."
    )
    format_parser.add_argument(
        "--folder",
        type=str,
        default=os.getcwd(),
        help="target folder or path, default is current folder, all protubuf files under it and its' subdirectories will be formatted."
    )
    format_parser.add_argument(
        "--indents",
        type=int,
        help="the number of indented spaces",
        default=4
    )
    format_parser.add_argument(
        "--top-comment",
        type=bool,
        default=False,
        help="format all comments as top comments(above the target line)"
    )
    format_parser.add_argument(
        "--align-by-equal-sign",
        type=bool,
        default=False,
        help="align the code by equal sign"
    )
    format_parser.add_argument(
        "--flatten",
        type=bool,
        default=False,
        help="flatten nested objects"
    )
    format_parser.add_argument(
        "--comment-max-length",
        type=int,
        default=999999,
        help="the max length of comment line, defalut is 999999."
    )
    usage = make_usage(parser)
    parser.usage = usage
    # patch format_help method to use custom usage message
    parser.format_help = types.MethodType(new_format_help, parser)
    # patch format_usage method to use custom usage message
    parser.format_usage = types.MethodType(new_format_usage, parser)
    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_usage()
        sys.exit(0)

    if args.command == 'format':
        root_path = os.getcwd()
        if args.files:
            proto_files = [os.path.join(root_path, file) for file in args.files]
        elif args.folder:
            proto_files = _get_proto_files(args.folder)
        else:
            proto_files = _get_proto_files(os.getcwd())

        proto_files = list(set(proto_files))  # remove duplicates
        for fp in proto_files:
            print(f"formatting {fp.replace(root_path, '')}")
            format_file(
                fp,
                indents=args.indents,
                top_comment=args.top_comment,
                align_by_equal_sign=args.align_by_equal_sign,
                flatten=args.flatten,
                comment_max_length=args.comment_max_length,
                new_fp=None
            )
        print("Done!")

    if args.command == 'view':
        args = parser.parse_args()
        fp = os.path.join(os.getcwd(), args.file)
        print(read_file(fp))



from enum import Enum


class WrapEnum(Enum):
    @classmethod
    def values(cls):
        """
        Get all enum values.
        :return: value list
        """
        return [v.value for name, v in vars(cls).items() if not name.startswith('_') and name != 'values']



import uuid
from attrdict import AttrDict
from enum import Enum
from .detector import Detector
from .comment import CommentParser
from .constant import Constant
from .proto_structures import EnumElement
from .proto_structures import Import
from .proto_structures import Message
from .proto_structures import MessageElement
from .proto_structures import Option
from .proto_structures import Package
from .proto_structures import Position
from .proto_structures import ProtoEnum
from .proto_structures import Service
from .proto_structures import ServiceElement
from .proto_structures import Syntax
from .proto_structures import Oneof
from .proto_structures import OneofElement
from .protobuf import Protobuf
from .util import remove_prefix, remove_suffix


class ObjectParser(Constant):

    def __init__(self):
        self.objects = []
        self.objects_dict = {}

        self.left_brace_stack = []
        self.right_brace_stack = []

        self.obj = None
        self.current_obj = None

    @classmethod
    def _get_obj_name(cls, line):
        parts = line.strip().split(' ')
        parts = list(filter(None, parts))
        name = parts[1]
        return name

    @classmethod
    def parse_obj_field(cls, line, top_comments):
        raise NotImplementedError("Element parse not implemented.")

    def parse_and_add(self, proto_obj: Protobuf, lines, top_comment_list):
        self.parse(lines, top_comment_list, True)
        proto_obj.objects.append(self.obj)

    def parse(self, lines):
        top_comments = Detector().get_top_comments(lines)
        if len(lines) == 0:
            return

        first_line = lines[0]
        if Detector()._is_object_start(first_line):
            self.left_brace_stack.append(self.BRACE_LEFT)

        type = Detector().get_type(first_line)
        if 'message' == type or 'enum' == type or 'service' == type or 'oneof' == type:
            name = self._get_obj_name(first_line)
            top_comments = CommentParser.create_comment(first_line, top_comments)
            obj = self.new_obj(type, name, top_comments)
            self.objects_dict[obj.id] = obj

            if self.obj:
                self.current_obj.elements.append(obj)
                obj.parent_id = self.current_obj.id
            else:
                self.obj = obj
                obj.parent_id = None
            self.current_obj = obj  # used as a pointer

            lines.pop(0)
            return self.parse(lines)
        elif 'enum_element' == type or 'message_element' == type or 'service_element' == type or 'oneof' == type:
            element = self.parse_element(type, first_line, top_comments)
            self.current_obj.elements.append(element)

            lines.pop(0)
            return self.parse(lines)
        else:
            if Detector()._is_object_end(first_line):
                self.right_brace_stack.append(self.BRACE_RIGHT)
                if self.current_obj.parent_id is None:  # root node, finish a root level object parse
                    self.objects.append(self.current_obj)
                    self.obj = None  # finshed a object and all netested inner objects(if has) parse.
                else:
                    self.current_obj = self.objects_dict[self.current_obj.parent_id]

                lines.pop(0)
                return self.parse(lines)

    @classmethod
    def new_obj(cls, type, name, comments=[]):
        obj_class = {
            'enum': ProtoEnum,
            'message': Message,
            'service': Service,
            'oneof': Oneof
        }[type]

        obj = obj_class(name=name, elements=[], comments=comments)
        obj.id = uuid.uuid4().hex

        return obj

    @classmethod
    def parse_element(cls, type, line, top_comments=[]):
        parse_method = {
            'enum_element': cls.parse_enum_element,
            'message_element': cls.parse_message_element,
            'service_element': cls.parse_service_element
        }[type]

        return parse_method(line, top_comments=top_comments)

    @classmethod
    def parse_enum_element(cls, line, top_comments=[]):
        # BAGGAGE_TYPE_CARRY_ON = 1;
        line = line.strip()
        equal_sign_index = line.index(cls.EQUAL_SIGN)
        semicolon_index = line.index(cls.SEMICOLON)
        str_before_equqal_sign = line[:equal_sign_index]
        parts = str_before_equqal_sign.split(' ')
        parts = list(filter(None, parts))
        value = line[equal_sign_index + 1:semicolon_index].strip()
        data = cls.get_number_and_rules(value)

        comments = CommentParser.create_comment(line, top_comments)
        return EnumElement(name=parts[0], number=data.number, rules=data.rules, comments=comments)

    @classmethod
    def parse_message_element(cls, line, top_comments=[]):
        # common.RequestContext  request_context = 1;
        # map<string, Project> projects = 3;
        # // x must be either "foo", "bar", or "baz"
        # string x = 1 [(validate.rules).string = {in: ["foo", "bar", "baz"]}];
        if 'map<' in line:
            return cls.make_map_element(line, top_comments)

        line = line.strip()
        equal_sign_index = line.index(cls.EQUAL_SIGN)
        semicolon_index = line.index(cls.SEMICOLON)
        str_before_equqal_sign = line[:equal_sign_index]
        parts = str_before_equqal_sign.split(' ')
        parts = list(filter(None, parts))
        value = line[equal_sign_index + 1:semicolon_index].strip()
        data = cls.get_number_and_rules(value)

        comments = CommentParser.create_comment(line, top_comments)
        if len(parts) == 2:
            return MessageElement(type=parts[0], name=parts[1], number=data.number, rules=data.rules,
                                  comments=comments)
        if len(parts) == 3:
            return MessageElement(label=parts[0], type=parts[1], name=parts[2], number=data.number, rules=data.rules,
                                  comments=comments)

        return None

    @classmethod
    def make_map_element(cls, line, top_comments=[]):
        # map<string, Project> projects = 3;
        right_bracket_index = line.index(cls.ANGLE_BRACKET_RIGHT)
        equal_sign_index = line.index(cls.EQUAL_SIGN)
        semicolon_index = line.index(cls.SEMICOLON)
        type = line[:right_bracket_index + 1]
        type = type.strip().replace(' ', '')
        type_parts = type.split(',')
        type = ', '.join(type_parts)
        name = line[right_bracket_index + 1:equal_sign_index]
        name = name.strip()
        number = line[equal_sign_index + 1:semicolon_index]
        number = number.strip()
        comments = CommentParser.create_comment(line, top_comments)

        return MessageElement(type=type, name=name, number=number, comments=comments)

    @classmethod
    def parse_service_element(cls, line, top_comments=[]):
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = line.strip().replace('(', '')
        line = line.replace(')', '')

        semicolon_index = line.index(cls.SEMICOLON)
        str_before_semicolon = line[:semicolon_index]
        parts = str_before_semicolon.split(' ')
        parts = list(filter(None, parts))
        comments = CommentParser.create_comment(line, top_comments)

        return ServiceElement(label=parts[0], name=parts[1], request=parts[2], response=parts[4], comments=comments)

    @classmethod
    def get_number_and_rules(self, value):
        number = value
        rules = ''
        if self.LEFT_SQUARE_BRACKET in value:
            left_brace_stack_index = value.index(self.LEFT_SQUARE_BRACKET)
            right_brace_stack_index = value.rindex(self.RIGHT_SQUARE_BRACKET)
            rules = value[left_brace_stack_index:right_brace_stack_index + 1]
            rules = rules.strip()
            number = value[:left_brace_stack_index]
            number = number.strip()

        return AttrDict({
            'number': number,
            'rules': rules
        })



def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


def remove_suffix(text, suffix):
    if text.endswith(suffix):
        return text[:-len(suffix)]
    return text


def is_word(line, target_index):
    return line[target_index + 1].lower() not in ' \n' or line[target_index - 1].lower() not in ' \n'


def find_word_start_index(line, target_index):
    index = target_index - 1
    while (index >= 0):
        if line[index].lower() in ' \n':
            return index
        index = index - 1
    return target_index


def find_word_end_index(line, target_index):
    index = target_index + 1
    while (True):
        if line[index].lower() in ' \n':
            return index
        index = index + 1
    return target_index


def to_lines(line, length):
    lines = []
    while (True):
        if len(line) <= length:
            lines.append(line)
            break

        if not is_word(line, length - 1):
            lines.append(line[:length - 1])
            sub_line = line[:length - 1]
            line = line[length - 1:].strip()
        else:
            index = find_word_start_index(line, length - 1)
            sub_line = line[:index]
            line = line[index:].strip()

        lines.append(sub_line)

    return lines


def read_file(file_path):
    with open(file_path) as f:
        return f.read()



from .proto_structures import Comment
from .constant import Constant
from .proto_structures import Position
from .util import remove_prefix, remove_suffix


class CommentParser(Constant):
    def __init__(self):
        self.end_line = 0

        self.multiple_comment_start_symbol_stack = []
        self.multiple_comment_end_symbol_stack = []

    def pick_up_comment(self, lines):
        comment_lines = []
        while lines:
            line = lines.pop(0).strip()
            # comment case: /* I'am a comment */
            if self._start_with_multiple_line_comment(line) and self._end_with_multiple_line_comment(line):
                comment_lines.append(line)
                continue

            if self._start_with_multiple_line_comment(line):
                self.multiple_comment_start_symbol_stack.append(self.MULTIPLE_COMENT_START_SYMBOL)
                comment_lines.append(line)
                continue

            if self.is_comment(line):
                comment_lines.append(line)

            if self._end_with_multiple_line_comment(line):
                self.multiple_comment_end_symbol_stack.append(self.MULTIPLE_COMENT_END_SYMBOL)
                continue

            if line and not self.is_comment(line):
                lines.insert(0, line)  # add back
                break

        return comment_lines

    @staticmethod
    def is_not_new_line(variable):
        return variable != '\n'

    def parse(self, comment_lines):
        comment_lines = self.remove_prefix_symble(comment_lines)

        comment_lines = list(filter(None, comment_lines))
        comment_lines = list(filter(self.is_not_new_line, comment_lines))

        return comment_lines

    def remove_prefix_symble(self, comment_lines):
        processed_lines = []
        for line in comment_lines:
            if line.strip().startswith(self.SINGLE_COMMENT_SYMBOL):
                # remove the single comment symble if have, E.g. // I'am a comment
                processed_lines.append(line.strip()[2:].strip())
            elif self._start_with_multiple_line_comment(line) or self._end_with_multiple_line_comment(line):
                # remove the multiple comment symble if have, E.g. /* I'am a comment */
                processed_lines.append(line.strip().replace(self.MULTIPLE_COMENT_START_SYMBOL, "").replace(
                    self.MULTIPLE_COMENT_END_SYMBOL, ''))
            elif (line.strip().startswith(self.STAR) or line.strip().startswith(
                    self.STAR * 2)) and not self._end_with_multiple_line_comment(line):
                # remove the multiple comment symble if have, E.g. * I'am a comment in multiple line comment
                # Example:
                # /*
                # **    Device information, including user agent, device type and ip address.
                # */
                #
                processed_lines.append(line.strip().replace(self.STAR, ""))
            else:
                processed_lines.append(line.strip())

        return processed_lines

    def _start_with_single_line_comment(self, line):
        return line.strip().startswith(self.SINGLE_COMMENT_SYMBOL)

    def _start_with_multiple_line_comment(self, line):
        return line.strip().startswith(self.MULTIPLE_COMENT_START_SYMBOL)

    def _end_with_multiple_line_comment(self, line):
        return line.strip().endswith(self.MULTIPLE_COMENT_END_SYMBOL)

    def _is_multiple_comment(self):
        if len(self.multiple_comment_start_symbol_stack) == 0:
            return False

        return len(self.multiple_comment_start_symbol_stack) > len(self.multiple_comment_end_symbol_stack)

    def is_comment(self, line):
        if self._is_multiple_comment():
            return True

        if self._start_with_single_line_comment(line):
            return True

        return False

    @staticmethod
    def parse_single_line_comment(line):
        comment = None
        if line.count(CommentParser.SINGLE_COMMENT_SYMBOL) > 0:
            start_index = line.index(CommentParser.SINGLE_COMMENT_SYMBOL)
            comment_str = line[start_index:].lstrip(CommentParser.SINGLE_COMMENT_SYMBOL).strip()
            comment = Comment(comment_str, Position.Right)
        return comment

    @staticmethod
    def create_top_comments(comments):
        result = []
        while comments:
            comment_lines = comments.pop(0)
            text = ''.join(comment_lines)
            text = text.strip()
            text = remove_prefix(text, CommentParser.SINGLE_COMMENT_SYMBOL)
            text = remove_prefix(text, CommentParser.MULTIPLE_COMENT_START_SYMBOL)
            text = remove_suffix(text, CommentParser.MULTIPLE_COMENT_END_SYMBOL)
            text = text.strip()
            result.append(Comment(text, Position.TOP))
        return result

    @staticmethod
    def create_comment(line, top_comment_list):
        comments = CommentParser.create_top_comments(top_comment_list)
        right_comment = CommentParser.parse_single_line_comment(line)
        if right_comment is not None:
            comments.append(right_comment)
        return comments
